@@ -158,11 +158,11 @@ branch_rec {
 
 unclosed_proc {
  mark:
-  Scheme_Closure_Data *d = (Scheme_Closure_Data *)p;
+  Scheme_Lambda *d = (Scheme_Lambda *)p;
 
   gcMARK2(d->name, gc);
-  gcMARK2(d->code, gc);
-  gcMARK2(d->closure_map, gc);
+  gcMARK2(d->body, gc);
+  gcMARK2(d->closure_map, gc); /* covers `ir_info` */
   gcMARK2(d->tl_map, gc);
 #ifdef MZ_USE_JIT
   gcMARK2(d->u.native_code, gc);
@@ -170,7 +170,7 @@ unclosed_proc {
 #endif
 
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Closure_Data));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Lambda));
 }
 
 let_value {
@@ -228,27 +228,50 @@ with_cont_mark {
   gcBYTES_TO_WORDS(sizeof(Scheme_With_Continuation_Mark));
 }
 
-comp_let_value {
+ir_local {
  mark:
-  Scheme_Compiled_Let_Value *c = (Scheme_Compiled_Let_Value *)p;
+  Scheme_IR_Local *var = (Scheme_IR_Local *)p;
 
-  gcMARK2(c->flags, gc);
-  gcMARK2(c->value, gc);
-  gcMARK2(c->body, gc);
-  gcMARK2(c->names, gc);
+  gcMARK2(var->name, gc);
+  switch (var->mode) {
+  case SCHEME_VAR_MODE_LETREC_CHECK:
+    gcMARK2(var->letrec_check.frame, gc);
+    break;
+  case SCHEME_VAR_MODE_OPTIMIZE:
+    gcMARK2(var->optimize.known_val, gc);
+    gcMARK2(var->optimize.transitive_uses, gc);
+    break;
+  case SCHEME_VAR_MODE_RESOLVE:
+    gcMARK2(var->resolve.lifted, gc);
+    break;
+  default:
+    break;
+  }
 
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Compiled_Let_Value));
+  gcBYTES_TO_WORDS(sizeof(Scheme_IR_Local));
+}
+
+ir_let_value {
+ mark:
+  Scheme_IR_Let_Value *c = (Scheme_IR_Let_Value *)p;
+
+  gcMARK2(c->value, gc);
+  gcMARK2(c->body, gc);
+  gcMARK2(c->vars, gc);
+
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_IR_Let_Value));
 }
 
 let_header {
  mark:
-  Scheme_Let_Header *h = (Scheme_Let_Header *)p;
+  Scheme_IR_Let_Header *h = (Scheme_IR_Let_Header *)p;
   
   gcMARK2(h->body, gc);
 
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Let_Header));
+  gcBYTES_TO_WORDS(sizeof(Scheme_IR_Let_Header));
 }
 
 set_bang {
@@ -308,14 +331,14 @@ closed_prim_proc {
 scm_closure {
   Scheme_Closure *c = (Scheme_Closure *)p;
   int closure_size = (c->code 
-                      ? ((Scheme_Closure_Data *)GC_resolve2(c->code, gc))->closure_size
+                      ? ((Scheme_Lambda *)GC_resolve2(c->code, gc))->closure_size
                       : 0);
 
  mark:
 
   int i = closure_size;
   START_MARK_ONLY;
-# define CLOSURE_DATA_TYPE Scheme_Closure_Data
+# define CLOSURE_DATA_TYPE Scheme_Lambda
 # include "mzclpf_decl.inc"
   END_MARK_ONLY;
 
@@ -363,6 +386,7 @@ cont_proc {
   gcMARK2(c->dw, gc);
   gcMARK2(c->prompt_tag, gc);
   gcMARK2(c->meta_continuation, gc);
+  gcMARK2(c->meta_continuation_src, gc);
   gcMARK2(c->common_dw, gc);
   gcMARK2(c->save_overflow, gc);
   gcMARK2(c->runstack_copied, gc);
@@ -482,7 +506,7 @@ bignum_obj {
   if (!SCHEME_BIGINLINE(b)) {
     gcMARK2(b->digits, gc);
   } else {
-    FIXUP_ONLY(b->digits = ((Small_Bignum *)GC_fixup_self(b))->v;)
+    b->digits = ((Small_Bignum *)GC_fixup_self(b))->v;
   }
 
  size:
@@ -564,11 +588,9 @@ bstring_obj {
 }
 
 symbol_obj {
-  Scheme_Symbol *s = (Scheme_Symbol *)p;
-
  mark:
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Symbol) + s->len + 1 - mzFLEX4_DELTA);
+  gcBYTES_TO_WORDS(sizeof(Scheme_Symbol) + ((Scheme_Symbol *)p)->len + 1 - mzFLEX4_DELTA);
 }
 
 cons_cell {
@@ -596,31 +618,25 @@ vector_obj {
 }
 
 fxvector_obj {
-  Scheme_Vector *vec = (Scheme_Vector *)p;
-
  mark:
  size:
   gcBYTES_TO_WORDS((sizeof(Scheme_Vector) 
-		    + ((vec->size - mzFLEX_DELTA) * sizeof(Scheme_Object *))));
+		    + ((((Scheme_Vector *)p)->size - mzFLEX_DELTA) * sizeof(Scheme_Object *))));
 }
 
 flvector_obj {
-  Scheme_Double_Vector *vec = (Scheme_Double_Vector *)p;
-
  mark:
  size:
   gcBYTES_TO_WORDS((sizeof(Scheme_Double_Vector) 
-		    + ((vec->size - mzFLEX_DELTA) * sizeof(double))));
+		    + ((((Scheme_Double_Vector *)p)->size - mzFLEX_DELTA) * sizeof(double))));
 }
 
 #ifdef MZ_LONG_DOUBLE
 extflvector_obj {
-  Scheme_Long_Double_Vector *vec = (Scheme_Long_Double_Vector *)p;
-
  mark:
  size:
   gcBYTES_TO_WORDS((sizeof(Scheme_Long_Double_Vector) 
-		    + ((vec->size - mzFLEX_DELTA) * sizeof(long double))));
+		    + ((((Scheme_Long_Double_Vector *)p)->size - mzFLEX_DELTA) * sizeof(long double))));
 }
 #endif
 
@@ -737,7 +753,8 @@ thread_val {
   gcMARK2(pr->returned_marks, gc);
   
   gcMARK2(pr->current_local_env, gc);
-  gcMARK2(pr->current_local_mark, gc);
+  gcMARK2(pr->current_local_scope, gc);
+  gcMARK2(pr->current_local_use_scope, gc);
   gcMARK2(pr->current_local_name, gc);
   gcMARK2(pr->current_local_modidx, gc);
   gcMARK2(pr->current_local_menv, gc);
@@ -766,8 +783,8 @@ thread_val {
   gcMARK2(pr->ku.k.p3, gc);
   gcMARK2(pr->ku.k.p4, gc);
   gcMARK2(pr->ku.k.p5, gc);
-  
-  gcMARK2(pr->list_stack, gc);
+
+  gcMARK2(pr->self_for_proc_chaperone, gc);
   
   gcMARK2(pr->kill_data, gc);
   gcMARK2(pr->private_kill_data, gc);
@@ -793,6 +810,16 @@ thread_val {
   gcMARK2(pr->mbox_first, gc);
   gcMARK2(pr->mbox_last, gc);
   gcMARK2(pr->mbox_sema, gc);
+
+  /* Follow msg_chain for an in-flight message like in place_async_channel_val */
+  {
+    Scheme_Object *cpr = pr->place_channel_msg_chain_in_flight;
+    while (cpr) {
+      gcMARK2(SCHEME_CAR(cpr), gc);
+      cpr = SCHEME_CDR(cpr);
+    }
+  }
+
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Thread));
 }
@@ -801,28 +828,30 @@ runstack_val {
   intptr_t *s = (intptr_t *)p;
  mark:
   void **a, **b;
-  a = (void **)s + 4 + s[2];
-  b = (void **)s + 4 + s[3];
+  a = (void **)s + 5 + s[2];
+  b = (void **)s + 5 + s[3];
   while (a < b) {
     gcMARK2(*a, gc);
     a++;
   }
- more:
- fixup:
+
+  START_MARK_ONLY;
   /* Zero out the part that we didn't mark, in case it becomes
      live later. */
-  a = (void **)s + 4;
-  b = (void **)s + 4 + s[2];
+  a = (void **)s + 5;
+  b = (void **)s + 5 + s[2];
   while (a < b) {
     *a = RUNSTACK_ZERO_VAL;
     a++;
   }
-  a = (void **)s + 4 + s[3];
-  b = (void **)s + 4 + (s[1] - 4);
+  a = (void **)s + 5 + s[3];
+  b = (void **)s + 5 + (s[1] - 5);
   while (a < b) {
     *a = RUNSTACK_ZERO_VAL;
     a++;
   }
+  END_MARK_ONLY;
+
  size:
   s[1];
 }
@@ -917,15 +946,15 @@ namespace_val {
   gcMARK2(e->guard_insp, gc);
   gcMARK2(e->access_insp, gc);
 
-  gcMARK2(e->rename_set, gc);
-  gcMARK2(e->temp_marked_names, gc);
-  gcMARK2(e->post_ex_rename_set, gc);
+  gcMARK2(e->stx_context, gc);
+  gcMARK2(e->tmp_bind_scope, gc);
 
   gcMARK2(e->syntax, gc);
   gcMARK2(e->exp_env, gc);
   gcMARK2(e->template_env, gc);
   gcMARK2(e->label_env, gc);
   gcMARK2(e->instance_env, gc);
+  gcMARK2(e->reader_env, gc);
 
   gcMARK2(e->shadowed_syntax, gc);
 
@@ -949,6 +978,8 @@ namespace_val {
 
   gcMARK2(e->weak_self_link, gc);
 
+  gcMARK2(e->binding_names, gc);
+
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Env));
 }
@@ -973,6 +1004,7 @@ compilation_top_val {
   Scheme_Compilation_Top *t = (Scheme_Compilation_Top *)p;
   gcMARK2(t->code, gc);
   gcMARK2(t->prefix, gc);
+  gcMARK2(t->binding_namess, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Compilation_Top));
@@ -987,7 +1019,7 @@ prefix_val {
  size:
   gcBYTES_TO_WORDS((sizeof(Scheme_Prefix) 
 		    + ((pf->num_slots-mzFLEX_DELTA) * sizeof(Scheme_Object *))
-                    + ((((pf->num_slots - (pf->num_stxes ? (pf->num_stxes+1) : 0)) + 31) / 32) 
+                    + ((((pf->num_slots - pf->num_stxes) + 31) / 32) 
                        * sizeof(int))));
 }
 
@@ -997,6 +1029,7 @@ resolve_prefix_val {
   gcMARK2(rp->toplevels, gc);
   gcMARK2(rp->stxes, gc);
   gcMARK2(rp->delay_info_rpair, gc);
+  gcMARK2(rp->src_insp_desc, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Resolve_Prefix));
@@ -1029,11 +1062,11 @@ stx_val {
   Scheme_Stx *stx = (Scheme_Stx *)p;
   gcMARK2(stx->val, gc);
   gcMARK2(stx->srcloc, gc);
-  gcMARK2(stx->wraps, gc);
+  gcMARK2(stx->scopes, gc);
+  gcMARK2(stx->u.to_propagate, gc);
+  gcMARK2(stx->shifts, gc);
   gcMARK2(stx->taints, gc);
   gcMARK2(stx->props, gc);
-  if (!(MZ_OPT_HASH_KEY(&(stx)->iso) & STX_SUBSTX_FLAG))
-    gcMARK2(stx->u.modinfo_cache, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Stx));
 }
@@ -1071,6 +1104,10 @@ module_val {
 
   gcMARK2(m->self_modidx, gc);
 
+  gcMARK2(m->binding_names, gc);
+  gcMARK2(m->et_binding_names, gc);
+  gcMARK2(m->other_binding_names, gc);
+
   gcMARK2(m->insp, gc);
 
   gcMARK2(m->lang_info, gc);
@@ -1078,6 +1115,7 @@ module_val {
   gcMARK2(m->hints, gc);
   gcMARK2(m->ii_src, gc);
   gcMARK2(m->super_bxs_info, gc);
+  gcMARK2(m->sub_iidx_ptrs, gc);
 
   gcMARK2(m->comp_prefix, gc);
   gcMARK2(m->prefix, gc);
@@ -1206,10 +1244,11 @@ mark_logger {
   Scheme_Logger *l = (Scheme_Logger *)p;
   gcMARK2(l->name, gc);
   gcMARK2(l->parent, gc);
-  gcMARK2(l->callback, gc);
-  gcMARK2(l->timestamp, gc);
+  gcMARK2(l->want_name_level_cache, gc);
+  gcMARK2(l->root_timestamp, gc);
   gcMARK2(l->syslog_level, gc);
   gcMARK2(l->stderr_level, gc);
+  gcMARK2(l->propagate_level, gc);
   gcMARK2(l->readers, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Logger));
@@ -1226,6 +1265,14 @@ mark_log_reader {
   gcBYTES_TO_WORDS(sizeof(Scheme_Log_Reader));
 }
 
+struct_proc_shape {
+ mark:
+  Scheme_Struct_Proc_Shape *s = (Scheme_Struct_Proc_Shape *)p;
+  gcMARK2(s->identity, gc);
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Struct_Proc_Shape));
+}
+
 END type;
 
 /**********************************************************************/
@@ -1240,30 +1287,35 @@ START compenv;
 
 mark_comp_env {
  mark:
-  Scheme_Full_Comp_Env *e = (Scheme_Full_Comp_Env *)p;
+  Scheme_Comp_Env *e = (Scheme_Comp_Env *)p;
 
-  gcMARK2(e->base.genv, gc);
-  gcMARK2(e->base.insp, gc);
-  gcMARK2(e->base.prefix, gc);
-  gcMARK2(e->base.next, gc);
-  gcMARK2(e->base.values, gc);
-  gcMARK2(e->base.renames, gc);
-  gcMARK2(e->base.uid, gc);
-  gcMARK2(e->base.uids, gc);
-  gcMARK2(e->base.dup_check, gc);
-  gcMARK2(e->base.intdef_name, gc);
-  gcMARK2(e->base.in_modidx, gc);
-  gcMARK2(e->base.skip_table, gc);
+  gcMARK2(e->genv, gc);
+  gcMARK2(e->insp, gc);
+  gcMARK2(e->prefix, gc);
+  gcMARK2(e->next, gc);
+  gcMARK2(e->scopes, gc);
+  gcMARK2(e->value_name, gc);
+  gcMARK2(e->observer, gc);
+  gcMARK2(e->binders, gc);
+  gcMARK2(e->bindings, gc);
+  gcMARK2(e->vals, gc);
+  gcMARK2(e->shadower_deltas, gc);
+  gcMARK2(e->vars, gc);
+  gcMARK2(e->dup_check, gc);
+  gcMARK2(e->intdef_name, gc);
+  gcMARK2(e->in_modidx, gc);
+  gcMARK2(e->skip_table, gc);
   
-  gcMARK2(e->data.const_names, gc);
-  gcMARK2(e->data.const_vals, gc);
-  gcMARK2(e->data.const_uids, gc);
-  gcMARK2(e->data.sealed, gc);
-  gcMARK2(e->data.use, gc);
-  gcMARK2(e->data.lifts, gc);
+  gcMARK2(e->use, gc);
+  gcMARK2(e->lifts, gc);
+  gcMARK2(e->bindings, gc);
+
+  gcMARK2(e->binding_namess, gc);
+
+  gcMARK2(e->expand_result_adjust_arg, gc);
 
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Full_Comp_Env));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Comp_Env));
 }
 
 END compenv;
@@ -1279,12 +1331,8 @@ mark_resolve_info {
   gcMARK2(i->prefix, gc);
   gcMARK2(i->stx_map, gc);
   gcMARK2(i->tl_map, gc);
-  gcMARK2(i->old_pos, gc);
-  gcMARK2(i->new_pos, gc);
-  gcMARK2(i->old_stx_pos, gc);
-  gcMARK2(i->flags, gc);
+  gcMARK2(i->redirects, gc);
   gcMARK2(i->lifts, gc);
-  gcMARK2(i->lifted, gc);
   gcMARK2(i->next, gc);
 
  size:
@@ -1295,10 +1343,20 @@ mark_unresolve_info {
  mark:
   Unresolve_Info *i = (Unresolve_Info *)p;
   
-  gcMARK2(i->flags, gc);
-  gcMARK2(i->depths, gc);
+  gcMARK2(i->vars, gc);
   gcMARK2(i->prefix, gc);
   gcMARK2(i->closures, gc);
+  gcMARK2(i->module, gc);
+  gcMARK2(i->comp_prefix, gc);
+  gcMARK2(i->new_toplevels, gc);
+  gcMARK2(i->from_modidx, gc);
+  gcMARK2(i->to_modidx, gc);
+  gcMARK2(i->opt_env, gc);
+  gcMARK2(i->opt_insp, gc);
+  gcMARK2(i->inline_variants, gc);
+  gcMARK2(i->toplevels, gc);
+  gcMARK2(i->definitions, gc);
+  gcMARK2(i->ref_lifts, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Unresolve_Info));
@@ -1335,7 +1393,6 @@ mark_letrec_check_frame {
   gcMARK2(frame->def, gc);
   gcMARK2(frame->next, gc);
   gcMARK2(frame->ref, gc);
-  gcMARK2(frame->head, gc);
   gcMARK2(frame->deferred_chain, gc);
 
  size:
@@ -1364,18 +1421,17 @@ mark_optimize_info {
  mark:
   Optimize_Info *i = (Optimize_Info *)p;
   
-  gcMARK2(i->stat_dists, gc);
-  gcMARK2(i->sd_depths, gc);
   gcMARK2(i->next, gc);
-  gcMARK2(i->use, gc);
   gcMARK2(i->consts, gc);
   gcMARK2(i->cp, gc);
+  gcMARK2(i->env, gc);
+  gcMARK2(i->insp, gc);
   gcMARK2(i->top_level_consts, gc);
-  gcMARK2(i->transitive_use, gc);
-  gcMARK2(i->transitive_use_len, gc);
+  gcMARK2(i->transitive_use_var, gc);
   gcMARK2(i->context, gc);
   gcMARK2(i->logger, gc);
   gcMARK2(i->types, gc);
+  gcMARK2(i->uses, gc);
 
  size:
   gcBYTES_TO_WORDS(sizeof(Optimize_Info));
@@ -1385,8 +1441,7 @@ mark_once_used {
  mark:
   Scheme_Once_Used *o = (Scheme_Once_Used *)p;
   gcMARK2(o->expr, gc);
-  gcMARK2(o->info, gc);
-  gcMARK2(o->next, gc);
+  gcMARK2(o->var, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Once_Used));
 }
@@ -1397,17 +1452,6 @@ END optimize;
 
 START eval;
 
-mark_comp_info {
- mark:
-  Scheme_Compile_Info *i = (Scheme_Compile_Info *)p;
-  
-  gcMARK2(i->value_name, gc);
-  gcMARK2(i->observer, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Compile_Info));
-}
-
 mark_saved_stack {
  mark:
   Scheme_Saved_Stack *saved = (Scheme_Saved_Stack *)p;
@@ -1442,18 +1486,6 @@ END validate;
 
 START fun;
 
-mark_closure_info {
- mark:
-  Closure_Info *i = (Closure_Info *)p;
-  
-  gcMARK2(i->local_flags, gc);
-  gcMARK2(i->base_closure_map, gc);
-  gcMARK2(i->local_type_map, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Closure_Info));
-}
-
 mark_dyn_wind_cell {
  mark:
   Scheme_Dynamic_Wind_List *l = (Scheme_Dynamic_Wind_List *)p;
@@ -1513,31 +1545,16 @@ END fun;
 START hash;
 
 hash_tree_val {
- mark:
   Scheme_Hash_Tree *ht = (Scheme_Hash_Tree *)p;
-
-  gcMARK2(ht->root, gc);
-
- size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Hash_Tree));
-}
-
-mark_avl_node {
+  int popcount = hamt_popcount(ht->bitmap);
  mark:
-  AVLNode *avl = (AVLNode *)p;
-
-  /* Short-circuit on NULL pointers, which are especially likely */
-  if (avl->left) {
-    gcMARK2(avl->left, gc);
+  int i;
+  for (i = ((SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL) ? 2 : 1) * popcount; i--; ) {
+    gcMARK2(ht->els[i], gc);
   }
-  if (avl->right) {
-    gcMARK2(avl->right, gc);
-  }
-  gcMARK2(avl->key, gc);
-  gcMARK2(avl->val, gc);
 
  size:
-  gcBYTES_TO_WORDS(sizeof(AVLNode));
+  gcBYTES_TO_WORDS(HASH_TREE_RECORD_SIZE(SCHEME_HASHTR_KIND(ht), popcount));
 }
 
 END hash;
@@ -1804,19 +1821,21 @@ mark_marshal_tables {
  mark:
   Scheme_Marshal_Tables *mt = (Scheme_Marshal_Tables *)p;
   gcMARK2(mt->symtab, gc);
-  gcMARK2(mt->rns, gc);
-  gcMARK2(mt->rn_refs, gc);
   gcMARK2(mt->st_refs, gc);
   gcMARK2(mt->st_ref_stack, gc);
-  gcMARK2(mt->reverse_map, gc);
-  gcMARK2(mt->same_map, gc);
-  gcMARK2(mt->shift_map, gc);
+  gcMARK2(mt->reachable_scopes, gc);
+  gcMARK2(mt->reachable_scope_stack, gc);
+  gcMARK2(mt->pending_reachable_ids, gc);
+  gcMARK2(mt->conditionally_reachable_scopes, gc);
+  gcMARK2(mt->intern_map, gc);
+  gcMARK2(mt->identity_map, gc);
   gcMARK2(mt->top_map, gc);
   gcMARK2(mt->key_map, gc);
   gcMARK2(mt->delay_map, gc);
   gcMARK2(mt->cdata_map, gc);
   gcMARK2(mt->rn_saved, gc);
   gcMARK2(mt->shared_offsets, gc);
+  gcMARK2(mt->path_cache, gc);
   gcMARK2(mt->sorted_keys, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(Scheme_Marshal_Tables));
@@ -2252,6 +2271,19 @@ END struct;
 
 START compile;
 
+mark_ir_lambda_info {
+ mark:
+  Scheme_IR_Lambda_Info *i = (Scheme_IR_Lambda_Info *)p;
+  
+  gcMARK2(i->base_closure, gc);
+  gcMARK2(i->vars, gc);
+  gcMARK2(i->arg_types, gc);
+  gcMARK2(i->arg_type_contributors, gc);
+
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_IR_Lambda_Info));
+}
+
 END compile;
 
 /**********************************************************************/
@@ -2272,11 +2304,13 @@ mark_cport {
   gcMARK2(cp->ht, gc);
   gcMARK2(cp->ut, gc);
   gcMARK2(cp->symtab, gc);
+  gcMARK2(cp->symtab_entries, gc);
   gcMARK2(cp->relto, gc);
   gcMARK2(cp->magic_sym, gc);
   gcMARK2(cp->magic_val, gc);
   gcMARK2(cp->shared_offsets, gc);
   gcMARK2(cp->delay_info, gc);
+  gcMARK2(cp->symtab_refs, gc);
  size:
   gcBYTES_TO_WORDS(sizeof(CPort));
 }
@@ -2309,6 +2343,7 @@ mark_delay_load {
   Scheme_Load_Delay *ld = (Scheme_Load_Delay *)p;
   gcMARK2(ld->path, gc);
   gcMARK2(ld->symtab, gc);
+  gcMARK2(ld->symtab_entries, gc);
   gcMARK2(ld->shared_offsets, gc);
   gcMARK2(ld->relto, gc);
   gcMARK2(ld->ut, gc);
@@ -2323,6 +2358,9 @@ mark_unmarshal_tables {
  mark:
   Scheme_Unmarshal_Tables *ut = (Scheme_Unmarshal_Tables *)p;
   gcMARK2(ut->rns, gc);
+  gcMARK2(ut->current_rns, gc);
+  gcMARK2(ut->multi_scope_pairs, gc);
+  gcMARK2(ut->current_multi_scope_pairs, gc);
   gcMARK2(ut->rp, gc);
   gcMARK2(ut->decoded, gc);
  size:
@@ -2392,35 +2430,6 @@ END string;
 
 START syntax;
 
-mark_rename_table {
- mark:
-  Module_Renames *rn = (Module_Renames *)p;
-  gcMARK2(rn->phase, gc);
-  gcMARK2(rn->ht, gc);
-  gcMARK2(rn->nomarshal_ht, gc);
-  gcMARK2(rn->unmarshal_info, gc);
-  gcMARK2(rn->shared_pes, gc);
-  gcMARK2(rn->set_identity, gc);
-  gcMARK2(rn->marked_names, gc);
-  gcMARK2(rn->free_id_renames, gc);
-  gcMARK2(rn->insp, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Module_Renames));
-}
-
-mark_rename_table_set {
- mark:
-  Module_Renames_Set *rns = (Module_Renames_Set *)p;
-  gcMARK2(rns->et, gc);
-  gcMARK2(rns->rt, gc);
-  gcMARK2(rns->other_phases, gc);
-  gcMARK2(rns->share_marked_names, gc);
-  gcMARK2(rns->set_identity, gc);
-  gcMARK2(rns->insp, gc);
- size:
-  gcBYTES_TO_WORDS(sizeof(Module_Renames_Set));
-}
-
 mark_srcloc {
  mark:
   Scheme_Stx_Srcloc *s = (Scheme_Stx_Srcloc *)p;
@@ -2429,27 +2438,38 @@ mark_srcloc {
   gcBYTES_TO_WORDS(sizeof(Scheme_Stx_Srcloc));
 }
 
-mark_wrapchunk {
-  Wrap_Chunk *wc = (Wrap_Chunk *)p;
+mark_scope {
+  Scheme_Scope *m = (Scheme_Scope *)p;
+  int for_multi = SCHEME_SCOPE_HAS_OWNER(m);
  mark:
-  int i;
-  for (i = wc->len; i--; ) {
-    gcMARK2(wc->a[i], gc);
+  gcMARK2(m->bindings, gc);
+  if (for_multi) {
+    gcMARK2(((Scheme_Scope_With_Owner *)m)->owner_multi_scope, gc);
+    gcMARK2(((Scheme_Scope_With_Owner *)m)->phase, gc);
   }
  size:
-  gcBYTES_TO_WORDS(sizeof(Wrap_Chunk) + ((wc->len - mzFLEX_DELTA) * sizeof(Scheme_Object *)));
+ (for_multi
+  ? gcBYTES_TO_WORDS(sizeof(Scheme_Scope_With_Owner))
+  : gcBYTES_TO_WORDS(sizeof(Scheme_Scope)));
 }
 
-lex_rib {
+mark_scope_table {
  mark:
-  Scheme_Lexical_Rib *rib = (Scheme_Lexical_Rib *)p;
-  gcMARK2(rib->rename, gc);
-  gcMARK2(rib->timestamp, gc);
-  gcMARK2(rib->sealed, gc);
-  gcMARK2(rib->mapped_names, gc);
-  gcMARK2(rib->next, gc);
+  Scheme_Scope_Table *m = (Scheme_Scope_Table *)p;
+  gcMARK2(m->simple_scopes, gc);
+  gcMARK2(m->multi_scopes, gc);
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Lexical_Rib));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Scope_Table));
+}
+
+mark_propagate_table {
+ mark:
+  Scheme_Propagate_Table *m = (Scheme_Propagate_Table *)p;
+  mark_scope_table_MARK(&m->st, gc);
+  gcMARK2(m->prev, gc);
+  gcMARK2(m->phase_shift, gc);
+ size:
+  gcBYTES_TO_WORDS(sizeof(Scheme_Propagate_Table));
 }
 
 END syntax;
@@ -2460,7 +2480,7 @@ START jit;
 
 native_closure {
   Scheme_Native_Closure *c = (Scheme_Native_Closure *)p;
-  int closure_size = ((Scheme_Native_Closure_Data *)GC_resolve2(c->code, gc))->closure_size;
+  int closure_size = ((Scheme_Native_Lambda *)GC_resolve2(c->code, gc))->closure_size;
 
   if (closure_size < 0) {
     closure_size = -(closure_size + 1);
@@ -2470,7 +2490,7 @@ native_closure {
   {
   int i = closure_size;
   START_MARK_ONLY;
-# define CLOSURE_DATA_TYPE Scheme_Native_Closure_Data
+# define CLOSURE_DATA_TYPE Scheme_Native_Lambda
 # include "mzclpf_decl.inc"
   END_MARK_ONLY;
 
@@ -2498,7 +2518,7 @@ mark_jit_state {
  mark:
   mz_jit_state *j = (mz_jit_state *)p;
   gcMARK2(j->mappings, gc);
-  gcMARK2(j->self_data, gc);
+  gcMARK2(j->self_lam, gc);
   gcMARK2(j->example_argv, gc);
   gcMARK2(j->nc, gc);
   gcMARK2(j->retaining_data, gc);
@@ -2510,7 +2530,7 @@ mark_jit_state {
 
 native_unclosed_proc {
  mark:
-  Scheme_Native_Closure_Data *d = (Scheme_Native_Closure_Data *)p;
+  Scheme_Native_Lambda *d = (Scheme_Native_Lambda *)p;
   int i;
 
   gcMARK2(d->u2.name, gc);
@@ -2523,20 +2543,21 @@ native_unclosed_proc {
     gcMARK2(d->u.arities, gc);
   }
   gcMARK2(d->tl_map, gc);
+  gcMARK2(d->eq_key, gc);
 
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Native_Closure_Data));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Native_Lambda));
 }
 
 native_unclosed_proc_plus_case {
  mark:
-  Scheme_Native_Closure_Data_Plus_Case *d = (Scheme_Native_Closure_Data_Plus_Case *)p;
+  Scheme_Native_Lambda_Plus_Case *d = (Scheme_Native_Lambda_Plus_Case *)p;
 
   native_unclosed_proc_MARK(p, gc);
   gcMARK2(d->case_lam, gc);
 
  size:
-  gcBYTES_TO_WORDS(sizeof(Scheme_Native_Closure_Data_Plus_Case));
+  gcBYTES_TO_WORDS(sizeof(Scheme_Native_Lambda_Plus_Case));
 }
 
 END jit;
